import asyncio
import logging
from pathlib import Path
from typing import Optional, Sequence

import click
from yarl import URL

from ..cli import main
from ..common import NEURO_EXTRAS_IMAGE
from ..image import _get_cluster_from_uri
from ..utils import get_neuro_client
from .archive import ArchiveType
from .operations import CopyOperation


SUPPORTED_ARCHIVE_TYPES = list(ArchiveType.get_extension_mapping().keys())
SUPPORTED_OBJECT_STORAGE_SCHEMES = {
    "AWS": "s3://",
    "GCS": "gs://",
    # originally, Azure's blob scheme is 'http(s)',
    # but we prepend 'azure+' to differentiate https vs azure
    "AZURE": "azure+https://",
    "HTTP": "http://",
    "HTTPS": "https://",
}

TEMP_UNPACK_DIR = Path.home() / ".neuro-tmp"

logger = logging.getLogger(__name__)


@main.group()
def data() -> None:
    """
    Data transfer operations.
    """
    pass


@data.command("transfer")
@click.argument("source")
@click.argument("destination")
def data_transfer(source: str, destination: str) -> None:
    """
    Copy data between storages on different clusters. \n
    Consider archiving dataset first for the sake of performance,
    if the dataset contains a lot (100k+) of small (< 100Kb each) files.
    """
    asyncio.run(_data_transfer(source, destination))


@data.command(
    "cp",
    help=(
        "Copy data between external object storage and cluster. "
        "Supported external object storage systems: "
        f"{list(SUPPORTED_OBJECT_STORAGE_SCHEMES.keys())}. "
        "Note: originally, Azure's blob storage scheme is 'http(s)', "
        "but we prepend 'azure+' to differenciate https vs azure"
    ),
)
@click.argument("source")
@click.argument("destination")
@click.option(
    "-x",
    "--extract",
    default=False,
    is_flag=True,
    help=(
        "Perform extraction of SOURCE into the DESTINATION directory. "
        "The archive type is derived from the file name. "
        f"Supported types: {', '.join(SUPPORTED_ARCHIVE_TYPES)}."
    ),
)
@click.option(
    "-c",
    "--compress",
    default=False,
    is_flag=True,
    help=(
        "Perform compression of SOURCE into the DESTINATION file. "
        "The archive type is derived from the file name. "
        f"Supported types: {', '.join(SUPPORTED_ARCHIVE_TYPES)}."
    ),
)
@click.option(
    "-v",
    "--volume",
    metavar="MOUNT",
    multiple=True,
    help=(
        "Mounts directory from vault into container. "
        "Use multiple options to mount more than one volume."
    ),
)
@click.option(
    "-e",
    "--env",
    metavar="VAR=VAL",
    multiple=True,
    help=(
        "Set environment variable in container. "
        "Use multiple options to define more than one variable."
    ),
)
@click.option(
    "-t",
    "--use-temp-dir",
    default=False,
    is_flag=True,
    help=(
        "DEPRECATED - need for temp dir is automatically detected, "
        "this flag will be removed in a future release. "
        "Download and extract / compress data (if needed) "
        "inside the temporary directory. "
        "Afterwards move resulted file(s) into the DESTINATION. "
        "NOTE: use it if 'storage:' is involved and "
        "extraction or compression is performed to speedup the process."
    ),
)
@click.option(
    "-s",
    "--preset",
    metavar="PRESET_NAME",
    help=("Preset name used for copy."),
)
@click.option(
    "-l",
    "--life_span",
    metavar="SECONDS",
    help=("Copy job life span in seconds."),
)
def data_cp(
    source: str,
    destination: str,
    extract: bool,
    compress: bool,
    volume: Sequence[str],
    env: Sequence[str],
    use_temp_dir: bool,
    preset: Optional[str] = None,
    life_span: Optional[int] = None,
) -> None:
    if use_temp_dir:
        logger.warn(
            "Flag -t/--use-temp-dir is DEPRECATED, "
            "and will be REMOVED in a future release. "
            "Need for temp dir is detected automatically from "
            "the source/destination type and compression/extraction flags"
        )
    try:

        async def run_copy() -> None:
            async with get_neuro_client() as client:
                operation = CopyOperation(
                    source=source,
                    destination=destination,
                    compress=compress,
                    extract=extract,
                    volumes=list(volume),
                    env=list(env),
                    life_span=life_span,
                    preset=preset,
                    client=client,
                )

                await operation.run()

        asyncio.run(run_copy())
    except Exception as e:
        logger.exception(e)
        raise click.ClickException(f"{e.__class__.__name__}: {e}")


# TODO: (A.K.) implement TransferOperation
async def _data_transfer(src_uri_str: str, dst_uri_str: str) -> None:
    async with get_neuro_client() as client:
        src_cluster_or_null = _get_cluster_from_uri(
            client, src_uri_str, scheme="storage"
        )
        dst_cluster = _get_cluster_from_uri(client, dst_uri_str, scheme="storage")

        if not src_cluster_or_null:
            src_cluster = client.cluster_name
        else:
            src_cluster = src_cluster_or_null

    if not dst_cluster:
        raise ValueError(
            f"Invalid destination path {dst_uri_str}: missing cluster name"
        )

    async with get_neuro_client(cluster=dst_cluster) as client:
        await client.storage.mkdir(URL("storage:"), parents=True, exist_ok=True)
        await _run_copy_container(src_cluster, src_uri_str, dst_uri_str)


async def _run_copy_container(
    src_cluster: str, src_uri_str: str, dst_uri_str: str
) -> None:
    args = [
        "neuro",
        "run",
        "-s",
        "cpu-small",
        "--pass-config",
        "-v",
        f"{dst_uri_str}:/storage:rw",
        "-e",
        f"NEURO_CLUSTER={src_cluster}",  # inside the job, switch neuro to 'src_cluster'
        "--life-span 10d",
        NEURO_EXTRAS_IMAGE,
        "--",
        f"neuro --show-traceback cp --progress -r -u -T {src_uri_str} /storage",
    ]
    cmd = " ".join(args)
    click.echo(f"Running '{cmd}'")
    subprocess = await asyncio.create_subprocess_shell(cmd)
    returncode = await subprocess.wait()
    if returncode != 0:
        raise click.ClickException("Unable to copy storage")



import asyncio
import logging
import random
import re
import sys
from math import ceil
from tempfile import TemporaryDirectory
from typing import Iterable, Iterator, List, Optional

import neuro_sdk
import pytest
from neuro_sdk import Client
from pytest_lazyfixture import lazy_fixture  # type: ignore
from tenacity import retry, stop_after_attempt, wait_random_exponential

from ..conftest import (
    TEST_DATA_COPY_CLOUD_TO_LOCAL,
    TEST_DATA_COPY_CLOUD_TO_PLATFORM,
    TEST_DATA_COPY_LOCAL_TO_CLOUD,
    TEST_DATA_COPY_LOCAL_TO_LOCAL,
    TEST_DATA_COPY_PLATFORM_TO_CLOUD,
)
from .cloud_to_local import generate_cloud_to_local_copy_configs
from .cloud_to_platform import generate_cloud_to_platform_copy_configs
from .local_to_cloud import generate_local_to_cloud_copy_configs
from .local_to_local import generate_local_to_local_copy_configs
from .platform_to_cloud import generate_platform_to_cloud_copy_configs
from .resources import CopyTestConfig
from .utils import _run_command


UUID4_PATTERN = r"[0-9a-f]{8}-[0-9a-f]{4}-[4][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}"
DISK_ID_PATTERN = rf"disk-{UUID4_PATTERN}"
DISK_ID_REGEX = re.compile(DISK_ID_PATTERN)

root_logger = logging.getLogger()
root_logger.setLevel(logging.INFO)
logger = logging.getLogger(__name__)


def get_all_data_copy_configs(
    random_subset: bool = False,
    fraction: float = 0.05,
) -> List[CopyTestConfig]:
    random.seed(2)
    configs: List[CopyTestConfig] = []

    def _add_configs(condition: bool, items: List[CopyTestConfig]) -> None:
        nonlocal configs
        if condition:
            if random_subset:
                k = ceil(len(items) * fraction)
                configs += random.sample(items, k=k)
            else:
                configs += items

    _add_configs(
        condition=TEST_DATA_COPY_LOCAL_TO_LOCAL,
        items=generate_local_to_local_copy_configs(),
    )
    _add_configs(
        condition=TEST_DATA_COPY_LOCAL_TO_CLOUD,
        items=generate_local_to_cloud_copy_configs(),
    )
    _add_configs(
        condition=TEST_DATA_COPY_CLOUD_TO_LOCAL,
        items=generate_cloud_to_local_copy_configs(),
    )
    _add_configs(
        condition=TEST_DATA_COPY_PLATFORM_TO_CLOUD,
        items=generate_platform_to_cloud_copy_configs(),
    )
    _add_configs(
        condition=TEST_DATA_COPY_CLOUD_TO_PLATFORM,
        items=generate_cloud_to_platform_copy_configs(),
    )

    # (A.K.) todo: make it into a fixture
    async def get_client() -> Client:
        return await neuro_sdk.get()

    client = asyncio.run(get_client())
    for c in configs:
        c.source.client = client
        c.destination.client = client
    return configs


def archive_types_are_compatible(
    source: Optional[str], destination: Optional[str]
) -> bool:
    groups = [
        {".tar.gz", ".tgz"},
        {".tar.bz2", ".tbz", ".tbz2"},
        {".tar"},
        {".zip"},
        {".gzip", ".gz"},
    ]
    return any(source in group and destination in group for group in groups)


@pytest.fixture(
    params=get_all_data_copy_configs(),
    ids=str,
)
def data_copy_config(request: pytest.FixtureRequest) -> Iterable[CopyTestConfig]:
    config: CopyTestConfig = request.param  # type: ignore
    yield config
    logger.info(f"Cleaning up destination after '{config.as_command(minimized=True)}'")
    config.destination.remove()


@pytest.fixture(
    params=get_all_data_copy_configs(random_subset=True),
    ids=str,
)
def data_copy_config_smoke(request: pytest.FixtureRequest) -> Iterable[CopyTestConfig]:
    config: CopyTestConfig = request.param  # type: ignore
    yield config
    logger.info(f"Cleaning up destination after '{config.as_command(minimized=True)}'")
    config.destination.remove()


@pytest.fixture(scope="session")
def tempdir_fixture() -> Iterator[str]:
    """Provide temporary folder, which will be removed upon test end

    WARNING: when used with pytest-xdist,
    folder will be created one per each worker!
    """
    with TemporaryDirectory() as tmpdir:
        logger.info(f"Created tempdir: {tmpdir}")
        yield tmpdir
    logger.info(f"Removed tempdir: {tmpdir}")


@pytest.mark.xfail(strict=False)  # TODO: remove when platform stabilizes
@pytest.mark.parametrize(
    argnames="config", argvalues=[lazy_fixture("data_copy_config")]
)
@pytest.mark.skipif(sys.platform == "win32", reason="tools don't work on Windows")
def test_data_copy(config: CopyTestConfig, tempdir_fixture: str, disk: str) -> None:
    config.source.patch_tempdir(tempdir_fixture)
    config.source.patch_disk(disk)
    config.destination.patch_tempdir(tempdir_fixture)
    config.destination.patch_disk(disk)
    _run_data_copy_test_from_config(config=config)


@pytest.mark.smoke
@pytest.mark.smoke_only
@pytest.mark.xfail(strict=False)  # TODO: remove when platform stabilizes
@pytest.mark.parametrize(
    argnames="config", argvalues=[lazy_fixture("data_copy_config_smoke")]
)
@pytest.mark.skipif(sys.platform == "win32", reason="tools don't work on Windows")
def test_data_copy_smoke(
    config: CopyTestConfig, tempdir_fixture: str, disk: str
) -> None:
    config.source.patch_tempdir(tempdir_fixture)
    config.source.patch_disk(disk)
    config.destination.patch_tempdir(tempdir_fixture)
    config.destination.patch_disk(disk)
    _run_data_copy_test_from_config(config=config)


@retry(stop=stop_after_attempt(5), wait=wait_random_exponential(min=10, max=60))
def _run_data_copy_test_from_config(config: CopyTestConfig) -> None:
    logger.info(f"Running test from {repr(config)}")
    destination = config.destination
    source = config.source
    extract_flag = config.extract_flag
    compress_flag = config.compress_flag
    supported_destination_schemas = {
        "local": {"local", "s3", "gs", "azure+https", "storage", "disk"},
        "http": {"local", "storage", "disk"},
        "https": {"local", "storage", "disk"},
        "s3": {"local", "storage", "disk"},
        "gs": {"local", "storage", "disk"},
        "azure+https": {"local", "storage", "disk"},
        "storage": {"local", "s3", "gs", "azure+https", "disk"},
        "disk": {
            "local",
            "s3",
            "gs",
            "azure+https",
            "storage",
        },
    }
    supported_resource_combination = (
        destination.schema in supported_destination_schemas[source.schema]
    )
    bad_extraction = not source.is_archive and extract_flag
    bad_compression = not destination.is_archive and compress_flag
    invalid_flag_combination = bad_extraction or bad_compression
    should_succeed = (
        supported_resource_combination
        and not invalid_flag_combination
        and not config.should_fail
    )
    reasons_to_fail = []
    if not supported_resource_combination:
        reasons_to_fail.append(
            f"copy from '{source.schema}' schema "
            f"to '{destination.schema}' schema is unsupported."
        )
    if invalid_flag_combination:
        reasons_to_fail.append(
            f"invalid combination of filetypes and compress/extract flags."
        )
    if config.should_fail and config.fail_reason:
        reasons_to_fail.append(config.fail_reason)
    returncode, stdout, stderr = _run_command(
        "neuro-extras", ["-vvv"] + config.as_command()
    )
    succeeded = returncode == 0
    verb = "should" if should_succeed else "should not"
    assert_fail_message = f"'{config.as_command(minimized=True)} {verb} succeed."
    if not should_succeed:
        assert_fail_message += f" Reasons: {'; '.join(reasons_to_fail)}"
    assert succeeded == should_succeed, assert_fail_message
    if not should_succeed:
        return
    if should_succeed:
        should_exist_message = (
            f"Destination {destination} " "should exist if copy succeeded"
        )
        assert destination.exists(), should_exist_message
    both_archives = source.is_archive and destination.is_archive
    compatible_archives = archive_types_are_compatible(
        source.file_extension, destination.file_extension
    )
    should_skip_compression = compress_flag and both_archives and compatible_archives
    if should_skip_compression:
        assert "Skipping compression step" in stdout, "Should skip compression"
    # TODO: add check for recompression



import asyncio
import itertools
import logging
import os
import sys
import threading
import warnings
from contextlib import asynccontextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from types import TracebackType
from typing import Any, AsyncIterator, Callable, Dict, List, Optional, Type

import neuro_sdk
from neuro_sdk import Client


logger = logging.getLogger(__name__)


class CLIRunner:
    """Utility class for running shell commands"""

    async def run_command(self, command: str, args: List[str]) -> None:
        """Execute command with args

        If resulting statuscode is non-zero, RuntimeError is thrown
        with stderr as a message.
        """
        logger.info(f"Executing: {[command] + args}")
        # logger.warn(f"Calling echo instead of actual command!")
        # process = await asyncio.create_subprocess_exec("echo", *([command] + args))

        process = await asyncio.create_subprocess_exec(command, *args)
        status_code = await process.wait()
        if status_code != 0:
            raise RuntimeError(process.stderr)


@asynccontextmanager
async def get_neuro_client(
    cluster: Optional[str] = None,
) -> AsyncIterator[neuro_sdk.Client]:
    client: neuro_sdk.Client = await neuro_sdk.get()
    cluster_orig: Optional[str] = None
    try:
        await client.__aenter__()

        cluster_orig = client.cluster_name
        if cluster is not None:
            if cluster != cluster_orig:
                logger.info(
                    f"Temporarily switching cluster: {cluster_orig} -> {cluster}"
                )
                await client.config.switch_cluster(cluster)  # typing: ignore
            else:
                logger.info(f"Already on cluster: {cluster}")
        yield client
    finally:
        # NOTE: bypass https://github.com/neuro-inc/platform-client-python/issues/1816
        try:
            await client.__aexit__(None, None, None)
        except asyncio.CancelledError:
            raise
        except BaseException as e:
            logger.warning(f"Ignoring exception during closing neuro client: {e}")

        if cluster is not None and cluster_orig is not None and cluster != cluster_orig:
            logger.info(f"Switching back cluster: {cluster} -> {cluster_orig}")
            try:
                await client.config.switch_cluster(cluster_orig)
            except asyncio.CancelledError:
                raise
            except BaseException:
                logger.error(
                    f"Could not switch back to cluster '{cluster_orig}'. Please "
                    f"run manually: 'neuro config switch-cluster {cluster_orig}'"
                )


if sys.platform != "win32":
    from asyncio.unix_events import AbstractChildWatcher

    _Callback = Callable[..., None]

    class ThreadedChildWatcher(AbstractChildWatcher):
        # Backport from Python 3.8

        """Threaded child watcher implementation.

        The watcher uses a thread per process
        for waiting for the process finish.

        It doesn't require subscription on POSIX signal
        but a thread creation is not free.

        The watcher has O(1) complexity, its performance doesn't depend
        on amount of spawn processes.
        """

        def __init__(self) -> None:
            self._pid_counter = itertools.count(0)
            self._threads: Dict[int, threading.Thread] = {}

        def close(self) -> None:
            pass

        def __enter__(self) -> "ThreadedChildWatcher":
            return self

        def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType],
        ) -> None:
            pass

        def __del__(self, _warn: Any = warnings.warn) -> None:
            threads = [
                thread for thread in list(self._threads.values()) if thread.is_alive()
            ]
            if threads:
                _warn(
                    f"{self.__class__} has registered but not finished child processes",
                    ResourceWarning,
                    source=self,
                )

        def add_child_handler(self, pid: int, callback: _Callback, *args: Any) -> None:
            loop = asyncio.get_event_loop()
            thread = threading.Thread(
                target=self._do_waitpid,
                name=f"waitpid-{next(self._pid_counter)}",
                args=(loop, pid, callback, args),
                daemon=True,
            )
            self._threads[pid] = thread
            thread.start()

        def remove_child_handler(self, pid: int) -> bool:
            # asyncio never calls remove_child_handler() !!!
            # The method is no-op but is implemented because
            # abstract base classe requires it
            return True

        def attach_loop(self, loop: Optional[asyncio.AbstractEventLoop]) -> None:
            pass

        def _do_waitpid(
            self,
            loop: asyncio.AbstractEventLoop,
            expected_pid: int,
            callback: _Callback,
            args: List[Any],
        ) -> None:
            assert expected_pid > 0

            try:
                pid, status = os.waitpid(expected_pid, 0)
            except ChildProcessError:
                # The child process is already reaped
                # (may happen if waitpid() is called elsewhere).
                pid = expected_pid
                returncode = 255
                logger.warning(
                    "Unknown child process pid %d, will report returncode 255", pid
                )
            else:
                returncode = _compute_returncode(status)
                if loop.get_debug():
                    logger.debug(
                        "process %s exited with returncode %s", expected_pid, returncode
                    )

            if loop.is_closed():
                logger.warning("Loop %r that handles pid %r is closed", loop, pid)
            else:
                loop.call_soon_threadsafe(callback, pid, returncode, *args)

            self._threads.pop(expected_pid)

    def _compute_returncode(status: int) -> int:
        if os.WIFSIGNALED(status):
            # The child process died because of a signal.
            return -os.WTERMSIG(status)
        elif os.WIFEXITED(status):
            # The child process exited (e.g sys.exit()).
            return os.WEXITSTATUS(status)
        else:
            # The child exited, but we don't understand its status.
            # This shouldn't happen, but if it does, let's just
            # return that status; perhaps that helps debug it.
            return status


def setup_child_watcher() -> None:
    if sys.platform == "win32":
        WindowsProactorEventLoopPolicy = asyncio.WindowsProactorEventLoopPolicy

        asyncio.set_event_loop_policy(WindowsProactorEventLoopPolicy())
    else:
        if sys.version_info < (3, 8):
            asyncio.set_child_watcher(ThreadedChildWatcher())


def select_job_preset(
    preset: Optional[str], client: Client, min_cpu: float = 2, min_mem: int = 4096
) -> Optional[str]:
    """
    Try to automatically select the best available preset for a task.
    Memory is specified in mebibytes.
    """
    good_presets = []
    good_presets_names = []
    # Build a shortlist of presets that could fit
    for cluster_preset_name, cluster_preset_info in client.presets.items():
        # Don't even try to use GPU machines for image builds
        # Also ignore scheduled presets (they don't work with schedule-timeout)
        # see https://github.com/neuro-inc/neuro-extras/issues/488
        if (
            cluster_preset_info.cpu >= min_cpu
            and cluster_preset_info.memory_mb >= min_mem
            and cluster_preset_info.gpu is None
            and not cluster_preset_info.scheduler_enabled
        ):
            good_presets.append((cluster_preset_name, cluster_preset_info))
            good_presets_names.append(cluster_preset_name)
    # Sort presets by cost - memory - cpu
    good_presets.sort(key=lambda p: (p[1].credits_per_hour, p[1].memory_mb, p[1].cpu))

    if preset is None:
        if len(good_presets) > 0:
            # Select the best preset
            preset_name = good_presets[0][0]
            logger.info(f"Automatically selected build preset {preset_name}")
            return preset_name
        else:
            # Fallback to the default preset selection mechanism by neuro sdk
            logger.warning(
                "No resource preset satisfying minimal hardware "
                "requirements was found in the cluster. "
                "The job might take long time to accomplish. "
                "Consider contacting your cluster manager or admin "
                "to adjust the cluster configuration"
            )
            return None
    else:
        if preset in good_presets_names:
            # If user asked for a preset, and it's a good one - let them use it
            return preset
        else:
            if len(good_presets) > 0:
                # We have a better preset
                logger.warning(
                    f"The selected resource preset {preset} does not "
                    f"satisfy recommended minimum hardware requirements. "
                    f"Consider using '{good_presets[0][0]}'"
                )
            elif client.presets[preset].gpu is None:
                # The message is only displayed if user selected a bad non-GPU preset
                logger.warning(
                    f"Selected resource preset {preset} does not satisfy "
                    f"minimal hardware requirements. "
                    "The job might take long time to accomplish. "
                    "Consider contacting your cluster manager or admin "
                    "to adjust the cluster configuration"
                )
            return preset


def get_default_preset(neuro_client: Client) -> str:
    """Get default preset name via Neu.ro client"""
    return next(iter(neuro_client.presets.keys()))


def provide_temp_dir(
    dir: Path = Path.home() / ".neuro-tmp",
) -> TemporaryDirectory:  # type: ignore
    """Provide temp directory"""
    dir.mkdir(exist_ok=True, parents=True)
    return TemporaryDirectory(dir=dir)



"""Module for data operations

Provides:
- CopyOperation
"""
import logging
from pathlib import Path
from typing import List, Optional, Tuple

from neuro_sdk import Client

from ..utils import provide_temp_dir
from .common import Copier, DataUrlType, Resource
from .local import CloudToLocalCopier, LocalToCloudCopier, LocalToLocalCopier
from .remote import RemoteCopier


logger = logging.getLogger(__name__)

# TODO: (A.K.) implement TransferOperation


class CopyOperation:
    """Abstraction of data copying between two locations
    with support for compression and extraction"""

    def __init__(
        self,
        source: str,
        destination: str,
        compress: bool,
        extract: bool,
        client: Client,
        volumes: Optional[List[str]] = None,
        env: Optional[List[str]] = None,
        life_span: Optional[float] = None,
        preset: Optional[str] = None,
    ) -> None:
        self.client = client
        self.source = Resource.parse(source, client=client)
        self.destination = Resource.parse(destination, client=client)
        self.compress = compress
        self.extract = extract
        self.volumes = volumes
        self.env = env
        self.life_span = life_span
        self.preset = preset
        self._ensure_can_execute()

    def _ensure_can_execute(self) -> None:
        """Raise exception if operation is unsupported

        Possible reasons:
        - at least one of the source or destination is of unsupported type
        - pair (source, destination) is explicitly forbidden
        """
        if not self.source.data_copy_supported:
            raise ValueError(f"Unsupported source: {self.source}")
        if not self.destination.data_copy_supported:
            raise ValueError(f"Unsupported destination: {self.destination}")
        source_type = DataUrlType.get_type(self.source.url)
        destination_type = DataUrlType.get_type(self.destination.url)
        is_forbidden_combination = any(
            (
                self.source.data_url_type == source
                and self.destination.data_url_type == destination
            )
            for (source, destination) in CopyOperation.get_forbidden_combinations()
        )
        if is_forbidden_combination:
            raise ValueError(
                f"Copy from {source_type.name} to "
                f"{destination_type.name} is unsupported. "
                "Please, reach us at https://github.com/neuro-inc/neuro-extras/issues "
                "describing your use case."
            )
        else:
            logger.debug(
                f"Copy from {source_type.name} to "
                f"{destination_type.name} is supported"
            )

    async def run(self) -> None:
        """Run copy operation.

        Uses appropriate copier instance, that supports source and destionation
        """

        with provide_temp_dir() as temp_dir:
            logger.debug("Resolving copier...")
            copier = _get_copier(
                source=self.source,
                destination=self.destination,
                compress=self.compress,
                extract=self.extract,
                temp_dir=Path(temp_dir),
                client=self.client,
                volumes=self.volumes,
                env=self.env,
                life_span=self.life_span,
                preset=self.preset,
            )
            logger.debug(f"Using {copier.__class__.__name__}")
            await copier.perform_copy()

    @staticmethod
    def get_forbidden_combinations() -> List[Tuple[DataUrlType, DataUrlType]]:
        """Get forbidden combinations of source and destination types"""
        return [
            (DataUrlType.CLOUD, DataUrlType.CLOUD),
            # TODO: (A.K.) implement platform-to-local and vice-versa
            # through neuro storage cp
            (DataUrlType.PLATFORM, DataUrlType.LOCAL_FS),
            (DataUrlType.LOCAL_FS, DataUrlType.PLATFORM),
            (DataUrlType.STORAGE, DataUrlType.STORAGE),
            (DataUrlType.DISK, DataUrlType.DISK),
            (DataUrlType.COPY_SUPPORTED, DataUrlType.WEB),
        ]


def _get_copier(
    source: Resource,
    destination: Resource,
    compress: bool,
    extract: bool,
    temp_dir: Path,
    client: Client,
    volumes: Optional[List[str]] = None,
    env: Optional[List[str]] = None,
    preset: Optional[str] = None,
    life_span: Optional[float] = None,
) -> Copier:
    """Resolve an instance of Copier, which is able to copy
    from source to destination with provided params"""
    source_type = DataUrlType.get_type(source.url)
    destination_type = DataUrlType.get_type(destination.url)
    if source_type == DataUrlType.LOCAL_FS and destination_type == DataUrlType.CLOUD:
        return LocalToCloudCopier(
            source=source,
            destination=destination,
            compress=compress,
            extract=extract,
            temp_dir=temp_dir,
        )
    elif source_type == DataUrlType.CLOUD and destination_type == DataUrlType.LOCAL_FS:
        return CloudToLocalCopier(
            source=source,
            destination=destination,
            compress=compress,
            extract=extract,
            temp_dir=temp_dir,
        )
    elif (
        source_type == DataUrlType.CLOUD
        and destination_type == DataUrlType.PLATFORM
        or source_type == DataUrlType.PLATFORM
        and destination_type == DataUrlType.CLOUD
        or source_type == DataUrlType.PLATFORM
        and destination_type == DataUrlType.PLATFORM
    ):
        return RemoteCopier(
            source=source,
            destination=destination,
            client=client,
            compress=compress,
            extract=extract,
            volumes=volumes,
            preset=preset,
            env=env,
            life_span=life_span,
        )
    elif source_type == DataUrlType.LOCAL_FS and destination_type.LOCAL_FS:
        return LocalToLocalCopier(
            source=source,
            destination=destination,
            compress=compress,
            extract=extract,
            temp_dir=temp_dir,
        )
    else:
        raise NotImplementedError(
            f"No copier found, that supports copy "
            f"from {source_type.name} to {destination_type.name}"
        )



"""Module for copying files from/to S3"""

from ..utils import CLIRunner
from .common import Copier, DataUrlType, Resource


class S3Copier(Copier, CLIRunner):
    """Copier, that is capable of copying to/from Amazon S3"""

    def _ensure_can_execute(self) -> None:
        if not (
            self.source.data_url_type == DataUrlType.LOCAL_FS
            and self.destination.data_url_type == DataUrlType.S3
            or self.source.data_url_type == DataUrlType.S3
            and self.destination.data_url_type == DataUrlType.LOCAL_FS
        ):
            raise ValueError(
                "Unsupported source and destination - "
                f"can only copy between {DataUrlType.S3.name} "
                f"and {DataUrlType.LOCAL_FS.name}"
            )

    async def perform_copy(self) -> Resource:
        """Perform copy through running aws cli and return the url to destinaton"""

        command = "aws"
        if self.source.as_str().endswith("/"):
            args = [
                "s3",
                "cp",
                "--recursive",
                self.source.as_str(),
                self.destination.as_str(),
            ]
        else:
            args = ["s3", "cp", self.source.as_str(), self.destination.as_str()]
        await self.run_command(command=command, args=args)
        return self.destination



"""Module for handling copy operations, that can be run locally

Contains:
- LocalToLocalCopier
- LocalToCloudCopier
- CloudToLocalCopier
"""
import logging
import os
import tempfile
from pathlib import Path
from typing import Type

from neuro_extras.data.fs import LocalFSCopier
from neuro_extras.data.web import WebCopier

from .archive import ArchiveType, compress, extract
from .azure import AzureCopier
from .common import Copier, DataUrlType, Resource, ensure_folder_exists
from .gcs import GCSCopier
from .s3 import S3Copier


logger = logging.getLogger(__name__)


class BaseLocalCopier(Copier):
    """Base class for copiers, which can be executed locally"""

    def __init__(
        self,
        source: Resource,
        destination: Resource,
        compress: bool = False,
        extract: bool = False,
        temp_dir: Path = Path(tempfile.gettempdir()),
    ) -> None:
        super().__init__(source=source, destination=destination)
        self.compress = compress
        self.extract = extract
        self.temp_dir = temp_dir

    @staticmethod
    def get_copier(
        source: Resource, destination: Resource, type: DataUrlType
    ) -> Copier:
        """Get copier of proper type to copy from"""
        destination_copier_mapping = {
            DataUrlType.S3: S3Copier,
            DataUrlType.AZURE: AzureCopier,
            DataUrlType.GCS: GCSCopier,
            DataUrlType.HTTP: WebCopier,
            DataUrlType.HTTPS: WebCopier,
            DataUrlType.LOCAL_FS: LocalFSCopier,
        }
        cls: Type[Copier] = destination_copier_mapping[type]
        return cls(source=source, destination=destination)

    def _can_skip_recompression(self) -> bool:
        """Check if both urls point to archives of same type"""
        if None in (self.source.filename, self.destination.filename):
            # at least one is a directory
            return False
        if ArchiveType.UNSUPPORTED in (
            self.source.archive_type,
            self.destination.archive_type,
        ):
            # at least one is not a supported archive
            return False
        return self.source.archive_type == self.destination.archive_type

    def _ensure_recompression_possible(self) -> None:
        """Raise error if recompression is not possible"""
        if None in (self.source.filename, self.destination.filename):
            # at least one is a directory
            raise ValueError("Recompression is unsupported for directories")
        if ArchiveType.UNSUPPORTED in (
            self.source.archive_type,
            self.destination.archive_type,
        ):
            raise ValueError("Recompression is not possible for unsupported archives")


class LocalToLocalCopier(BaseLocalCopier):
    """Copier, that can copy data from and to local storage

    Supports compression and extraction.
    """

    def _ensure_can_execute(self) -> None:
        if not (
            self.source.data_url_type == DataUrlType.LOCAL_FS
            and self.destination.data_url_type == DataUrlType.LOCAL_FS
        ):
            raise ValueError(
                f"Can only copy from {DataUrlType.LOCAL_FS.name} "
                f"to {DataUrlType.LOCAL_FS.name}"
            )

    async def _recompress(self) -> Resource:
        self._ensure_recompression_possible()
        if self._can_skip_recompression():
            logger.info(
                "Skipping compression step - "
                "source and destination are archives of the same type"
            )
            return await self._copy()
        temp_extraction_destination = self.temp_dir / "extracted"
        extracted_folder = await extract(
            source=self.source,
            destination=Resource.from_path(temp_extraction_destination),
        )

        new_archive = await compress(
            source=extracted_folder, destination=self.destination
        )
        return new_archive

    async def _extract(self) -> Resource:
        if self.source.filename is None:
            raise ValueError(f"Can't infer archive type from source {self.source}")
        extracted_folder = await extract(
            source=self.source,
            destination=self.destination,
        )
        return extracted_folder

    async def _compress(self) -> Resource:
        if self.destination.filename is None:
            raise ValueError(
                f"Can't infer archive type from destination {self.destination}"
            )
        compressed_file = await compress(
            source=self.source,
            destination=self.destination,
        )
        return compressed_file

    async def _copy(self) -> Resource:
        copier_implementation = BaseLocalCopier.get_copier(
            source=self.source, destination=self.destination, type=DataUrlType.LOCAL_FS
        )
        return await copier_implementation.perform_copy()

    async def perform_copy(self) -> Resource:
        """Perform copy from local fs to local fs.

        Delegates copy implementation to appropriate LocalFSCopier.
        Uses ArchiveManager to handle compression/extraction.
        """
        ensure_folder_exists(self.destination)
        if self.extract and self.compress:
            return await self._recompress()
        elif self.extract:
            return await self._extract()
        elif self.compress:
            return await self._compress()
        else:
            return await self._copy()


class LocalToCloudCopier(BaseLocalCopier):
    """Copier, that can copy data from local storage to a storage bucket

    Supports compression and extraction (temp_dir is used to store intermediate results)
    """

    def _ensure_can_execute(self) -> None:
        if not (
            self.source.data_url_type == DataUrlType.LOCAL_FS
            and self.destination.data_url_type == DataUrlType.CLOUD
        ):
            raise ValueError(
                f"Can only copy from {DataUrlType.LOCAL_FS.name} "
                f"to {DataUrlType.CLOUD.name}"
            )

    async def _recompress_and_copy(self) -> Resource:
        self._ensure_recompression_possible()
        if self._can_skip_recompression():
            logger.info(
                "Skipping compression step - "
                "source and destination are archives of the same type"
            )
            return await self._copy()
        temp_archive = (
            self.temp_dir / "recompressed" / self.destination.filename  # type: ignore
        )
        local_copier = LocalToLocalCopier(
            source=self.source,
            destination=Resource.from_path(temp_archive),
            extract=True,
            compress=True,
            temp_dir=self.temp_dir,
        )
        logger.debug(f"Using {local_copier} to perform recompression locally")
        copy_source = await local_copier.perform_copy()
        copier_implementation = BaseLocalCopier.get_copier(
            source=copy_source,
            destination=self.destination,
            type=self.destination.data_url_type,
        )
        return await copier_implementation.perform_copy()

    async def _extract_and_copy(self) -> Resource:
        if self.source.filename is None:
            raise ValueError(f"Can't infer archive type from source {self.source}")
        extracted_folder = await extract(
            source=self.source, destination=Resource.from_path(self.temp_dir)
        )
        copy_source = str(extracted_folder) + os.sep
        copier_implementation = BaseLocalCopier.get_copier(
            source=Resource.from_str(copy_source),
            destination=self.destination,
            type=self.destination.data_url_type,
        )
        return await copier_implementation.perform_copy()

    async def _compress_and_copy(self) -> Resource:
        if self.destination.filename is None:
            raise ValueError(
                f"Can't infer archive type from destination {self.destination}"
            )
        compressed_file = await compress(
            source=self.source,
            destination=Resource.from_path(self.temp_dir / self.destination.filename),
        )
        copier_implementation = BaseLocalCopier.get_copier(
            source=compressed_file,
            destination=self.destination,
            type=self.destination.data_url_type,
        )
        return await copier_implementation.perform_copy()

    async def _copy(self) -> Resource:
        copier_implementation = BaseLocalCopier.get_copier(
            source=self.source,
            destination=self.destination,
            type=self.destination.data_url_type,
        )
        return await copier_implementation.perform_copy()

    async def perform_copy(self) -> Resource:
        """Perform copy from local to cloud.

        Delegates copy implementation to appropriate Copier (S3, Azure, GCS, Web).
        Uses ArchiveManager to handle compression/extraction.
        """
        if self.extract and self.compress:
            return await self._recompress_and_copy()
        elif self.extract:
            return await self._extract_and_copy()
        elif self.compress:
            return await self._compress_and_copy()
        else:
            return await self._copy()


class CloudToLocalCopier(BaseLocalCopier):
    """Copier, that can copy data from cloud storage to local fs

    Supports compression and extraction (temp_dir is used to store intermediate results)
    """

    def _ensure_can_execute(self) -> None:
        if not (
            self.source.data_url_type == DataUrlType.CLOUD
            and self.destination.data_url_type == DataUrlType.LOCAL_FS
        ):
            raise ValueError(
                "Unsupported source and destination - "
                f"can only copy between {DataUrlType.CLOUD.name} "
                f"and {DataUrlType.LOCAL_FS.name}"
            )

    async def _copy_and_recompress(self) -> Resource:
        self._ensure_recompression_possible()
        if self._can_skip_recompression():
            logger.info(
                "Skipping compression step - "
                "source and destination are archives of the same type"
            )
            return await self._copy()
        temp_source_archive: Path = self.temp_dir / self.source.filename  # type: ignore
        copier_implementation = BaseLocalCopier.get_copier(
            source=self.source,
            destination=Resource.from_path(temp_source_archive),
            type=self.source.data_url_type,
        )
        temp_location = await copier_implementation.perform_copy()
        local_copier = LocalToLocalCopier(
            source=temp_location,
            destination=self.destination,
            extract=True,
            compress=True,
            temp_dir=self.temp_dir,
        )
        logger.debug(f"Using {local_copier} to perform recompression locally")
        return await local_copier.perform_copy()

    async def _copy_and_extract(self) -> Resource:
        if self.source.filename is None:
            raise ValueError(f"Can't infer archive type from source {self.source}")
        temp_archive = self.temp_dir / self.source.filename
        copier_implementation = BaseLocalCopier.get_copier(
            source=self.source,
            destination=Resource.from_path(temp_archive),
            type=self.source.data_url_type,
        )
        archive = await copier_implementation.perform_copy()
        extraction_result = await extract(source=archive, destination=self.destination)
        return extraction_result

    async def _copy_and_compress(self) -> Resource:
        if self.destination.filename is None:
            raise ValueError(
                f"Can't infer archive type from destination {self.destination}"
            )
        self.temp_dir.mkdir(exist_ok=True, parents=True)
        if self.source.filename:
            destination_path = str(self.temp_dir / self.source.filename)
        else:
            (self.temp_dir / "source").mkdir(exist_ok=True, parents=True)
            destination_path = str(self.temp_dir / "source") + os.sep
        copier_implementation = BaseLocalCopier.get_copier(
            source=self.source,
            destination=Resource.from_str(destination_path),
            type=self.source.data_url_type,
        )
        directory = await copier_implementation.perform_copy()
        compression_result = await compress(
            source=directory, destination=self.destination
        )
        return compression_result

    async def _copy(self) -> Resource:
        copier_implementation = BaseLocalCopier.get_copier(
            source=self.source,
            destination=self.destination,
            type=self.source.data_url_type,
        )
        return await copier_implementation.perform_copy()

    async def perform_copy(self) -> Resource:
        ensure_folder_exists(self.destination)
        if self.extract and self.compress:
            return await self._copy_and_recompress()
        elif self.extract:
            return await self._copy_and_extract()
        elif self.compress:
            return await self._copy_and_compress()
        else:
            return await self._copy()



import logging
from typing import List, Tuple

from ..conftest import run_cli


logger = logging.getLogger(__name__)


def _run_command(cmd: str, args: List[str]) -> Tuple[int, str, str]:
    logger.info(f"Running: {[cmd] + args}")
    result = run_cli([cmd] + args)
    return result.returncode, result.stdout, result.stderr



"""Module for copying files from/to Azure"""
import logging
import os
from urllib import parse

from yarl import URL

from ..utils import CLIRunner
from .common import Copier, DataUrlType, Resource


logger = logging.getLogger(__name__)


class AzureCopier(Copier, CLIRunner):
    """Copier, that is capable of copying to/from Azure storage"""

    def _ensure_can_execute(self) -> None:
        if not (
            self.source.data_url_type == DataUrlType.LOCAL_FS
            and self.destination.data_url_type == DataUrlType.AZURE
            or self.source.data_url_type == DataUrlType.AZURE
            and self.destination.data_url_type == DataUrlType.LOCAL_FS
        ):
            raise ValueError(
                "Unsupported source and destination - "
                f"can only copy between {DataUrlType.AZURE.name} "
                f"and {DataUrlType.LOCAL_FS.name}"
            )

    async def perform_copy(self) -> Resource:
        """Perform copy from self.source into self.destination through rclone
        and return url to the copied resource"""

        sas_url_source = (
            self.source
            if self.source.data_url_type == DataUrlType.AZURE
            else self.destination
        )
        sas_url = _build_sas_url(sas_url_source.url)
        source = _patch_azure_url_for_rclone(self.source.url)
        destination = _patch_azure_url_for_rclone(self.destination.url)
        command = "rclone"
        args = ["copyto", "-v", "--azureblob-sas-url", sas_url, source, destination]
        await self.run_command(command=command, args=args)
        return self.destination


def _build_sas_url(azure_url: URL) -> str:
    """
    In order to build SAS URL we replace original URL scheme with HTTPS,
    remove everything from path except bucket name and append SAS token as a query
    """
    sas_token = os.getenv("AZURE_SAS_TOKEN", "")
    if not sas_token:
        logger.warning("AZURE_SAS_TOKEN env is not provided")
    quoted_url = (
        azure_url.with_scheme("https")
        .with_path("/".join(azure_url.path.split("/")[:2]))
        .with_query(sas_token)
    )
    # with_query performs urlencode of sas_token, which breaks the token,
    # so we urldecode the resulting url
    sas_url = parse.unquote(str(quoted_url))
    logger.debug(f"SAS URL: {sas_url}")
    return sas_url


def _patch_azure_url_for_rclone(url: URL) -> str:
    """Replace host part with :azureblob: if the url is of azure type"""
    if url.scheme == "azure+https":
        return f":azureblob:{url.path}"
    else:
        return str(url)



"""Module for copying files by running neu.ro jobs"""
import logging
from dataclasses import dataclass, replace
from typing import List, Mapping, Optional, Tuple

from neuro_cli.utils import resolve_disk
from neuro_sdk import Client, DiskVolume, RemoteImage, SecretFile, Volume
from yarl import URL

from ..common import EX_OK, NEURO_EXTRAS_IMAGE, _attach_job_stdout
from ..utils import get_default_preset, select_job_preset
from .common import Copier, DataUrlType, Resource


logger = logging.getLogger(__name__)


@dataclass
class RemoteJobConfig:
    """Arguments, passed to `neuro_sdk.Client.jobs.start()`"""

    image: RemoteImage
    command: str
    env: Optional[Mapping[str, str]]
    secret_env: Optional[Mapping[str, URL]]
    volumes: List[Volume]
    secret_files: List[SecretFile]
    disk_volumes: List[DiskVolume]
    preset_name: str
    life_span: Optional[float]
    pass_config: bool

    @staticmethod
    def create(
        source: Resource,
        destination: Resource,
        neuro_client: Client,
        compress: bool = False,
        extract: bool = False,
        volumes: Optional[List[str]] = None,
        env: Optional[List[str]] = None,
        preset: Optional[str] = None,
        life_span: Optional[float] = None,
    ) -> "RemoteJobConfig":
        """Create `RemoteJobConfig` for a neu.ro copy job.

        Copy job will copy data from `source` to `destination`"""
        image = neuro_client.parse.remote_image(NEURO_EXTRAS_IMAGE)

        (patched_source, patched_destination, data_mounts) = _map_into_volumes(
            source=source,
            destination=destination,
        )
        logger.debug(f"Patched {source} into {patched_source}")
        logger.debug(f"Patched {destination} into {patched_destination}")
        logger.debug(f"Created job mountpoints: {data_mounts}")
        command = _build_data_copy_command(
            source=patched_source,
            destination=patched_destination,
            extract=extract,
            compress=compress,
        )
        all_volumes = volumes + data_mounts if volumes else data_mounts
        env_parse_result = neuro_client.parse.envs(env if env else [])
        volume_parse_result = neuro_client.parse.volumes(all_volumes)
        preset_name = select_job_preset(
            preset=preset, client=neuro_client, min_cpu=1, min_mem=2048
        ) or get_default_preset(neuro_client)
        return RemoteJobConfig(
            image=image,
            command=command,
            env=env_parse_result.env,
            secret_env=env_parse_result.secret_env,
            volumes=list(volume_parse_result.volumes),
            disk_volumes=list(volume_parse_result.disk_volumes),
            secret_files=list(volume_parse_result.secret_files),
            preset_name=preset_name,
            life_span=life_span,
            pass_config=True,
        )


class RemoteCopier(Copier):
    """Copier, that creates a job on neu.ro platform.
    Can copy data between neu.ro storage: or disk: and cloud storage."""

    def __init__(
        self,
        source: Resource,
        destination: Resource,
        client: Client,
        compress: bool = False,
        extract: bool = False,
        volumes: Optional[List[str]] = None,
        env: Optional[List[str]] = None,
        preset: Optional[str] = None,
        life_span: Optional[float] = None,
    ) -> None:
        super().__init__(source, destination)
        self.neuro_client = client
        self.job_config = RemoteJobConfig.create(
            source=source,
            destination=destination,
            neuro_client=client,
            compress=compress,
            extract=extract,
            volumes=volumes,
            env=env,
            preset=preset,
            life_span=life_span,
        )

    def _ensure_can_execute(self) -> None:
        if not (
            self.source.data_url_type == DataUrlType.PLATFORM
            and self.destination.data_url_type == DataUrlType.CLOUD
            or self.source.data_url_type == DataUrlType.CLOUD
            and self.destination.data_url_type == DataUrlType.PLATFORM
        ):
            raise ValueError(
                f"Can only copy between {DataUrlType.PLATFORM} and {DataUrlType.CLOUD}"
            )

    async def perform_copy(self) -> Resource:
        # neuro_client.jobs.start accepts disk URIs with IDs only
        resolved_disks = []
        for disk in self.job_config.disk_volumes:
            disk_id = await resolve_disk(disk.disk_uri, client=self.neuro_client)
            disk_with_id = replace(disk, disk_uri=disk.disk_uri / f"../{disk_id}")
            resolved_disks.append(disk_with_id)

        self.job_config = replace(self.job_config, disk_volumes=resolved_disks)

        logger.info(f"Starting job from config: {self.job_config}")
        job = await self.neuro_client.jobs.start(
            image=self.job_config.image,
            command=self.job_config.command,
            env=self.job_config.env,
            secret_env=self.job_config.secret_env,
            volumes=self.job_config.volumes,
            secret_files=self.job_config.secret_files,
            disk_volumes=self.job_config.disk_volumes,
            preset_name=self.job_config.preset_name,
            life_span=self.job_config.life_span,
            pass_config=True,
        )
        logger.info(f"Started job {job.id}")
        exit_code = await _attach_job_stdout(job, self.neuro_client, name="copy")
        if exit_code == EX_OK:
            logger.info("Copy job finished")
        else:
            raise RuntimeError(f"Copy job failed: error code {exit_code}")

        return self.destination


def _map_into_volumes(
    source: Resource,
    destination: Resource,
    source_storage_mount_prefix: str = "/var/storage/source",
    destination_storage_mount_prefix: str = "/var/storage/destination",
    source_disk_mount_prefix: str = "/var/disk/source",
    destination_disk_mount_prefix: str = "/var/disk/destination",
) -> Tuple[str, str, List[str]]:
    """Map urls for platform storage into volume mounts.

    Returns (patched_source: str, patched_destination: str, volumes: List[str]),
    where patched_source and patched_destination are mount points for
    source and destination if they belong to platform storage
    and the same urls otherwise.
    """
    new_source, source_mounts = _map_resource_into_volume(
        resource=source,
        storage_mount_prefix=source_storage_mount_prefix,
        disk_mount_prefix=source_disk_mount_prefix,
    )

    new_destination, destination_mounts = _map_resource_into_volume(
        resource=destination,
        storage_mount_prefix=destination_storage_mount_prefix,
        disk_mount_prefix=destination_disk_mount_prefix,
        mount_files=False,
    )

    return new_source, new_destination, source_mounts + destination_mounts


def _map_resource_into_volume(
    resource: Resource,
    storage_mount_prefix: str,
    disk_mount_prefix: str,
    mount_files: bool = False,
    mount_mode: str = "rw",
) -> Tuple[str, List[str]]:
    """Map storage or disk resource into volume specification string
    and patch such url into a local path inside a container.

    storage: urls are mounted into storage_mount_prefix.
        If the url points to a file and mount_files is True,
        it is directly mounted as-is; is mount_files is False,
        parent directory is mounted instead (useful for
        mounting destination, as empy volumes are mounted as directories).

    disk: urls are mounted into disk_mount_prefix.
        If the path on disk is provided, patch url into that
        subfolder of disk_mount prefix.

    Other urls are left as-is.
    """
    volumes = []
    resource = resource.strip_mount_mode_flag()
    if resource.data_url_type == DataUrlType.STORAGE:
        if resource.filename:
            if mount_files:
                resource_url = resource.as_str()
                mountpoint = f"{storage_mount_prefix}/{resource.filename}"
            else:
                resource_url = resource.strip_filename().as_str()
                mountpoint = f"{storage_mount_prefix}/"
            new_url = f"{storage_mount_prefix}/{resource.filename}"
        else:
            resource_url = resource.as_str()
            mountpoint = f"{storage_mount_prefix}/"
            new_url = mountpoint
        volumes.append(f"{resource_url}:{mountpoint}:{mount_mode}")
    elif resource.data_url_type == DataUrlType.DISK:
        disk_full_id, path_on_disk = resource.disk_id_and_path
        mount_mode = resource.mode_flag if resource.mode_flag else mount_mode
        logger.debug(
            f"Parsed disk url {resource.url} into disk_id: {disk_full_id} "
            f"path_on_disk: {path_on_disk}, mode: {mount_mode}"
        )
        mountpoint = f"{disk_mount_prefix}/"
        new_url = f"{disk_mount_prefix}{path_on_disk}" if path_on_disk else mountpoint
        volumes.append(f"{disk_full_id}:{mountpoint}:{mount_mode}")
    else:
        new_url = resource.as_str()
    return new_url, volumes


def _build_data_copy_command(
    source: str, destination: str, extract: bool, compress: bool
) -> str:
    """Build a neuro-extras data cp command"""
    command_prefix = ["neuro-extras", "-v", "data", "cp"]
    args = [source, destination]
    flags = []
    if compress:
        flags.append("-c")
    if extract:
        flags.append("-x")
    full_command = command_prefix + flags + args
    return " ".join(full_command)



import uuid
from typing import Callable, ContextManager

import pytest

from ..conftest import CLIRunner


@pytest.mark.smoke
@pytest.mark.serial
def test_data_transfer(
    cli_runner: CLIRunner,
    current_user: str,
    switch_cluster: Callable[[str], ContextManager[None]],
    src_cluster: str,
    dst_cluster: str,
) -> None:
    # Note: data-transfer runs copying job on dst_cluster and
    # we pushed test image to src_cluster, so it should be a target cluster
    src_cluster, dst_cluster = dst_cluster, src_cluster

    with switch_cluster(src_cluster):
        result = cli_runner(["neuro-extras", "init-aliases"])
        assert result.returncode == 0, result

        src_path = (
            f"storage://{src_cluster}/{current_user}/copy-src/{str(uuid.uuid4())}"
        )
        result = cli_runner(["neuro", "mkdir", "-p", src_path])
        assert result.returncode == 0, result

        dst_path = (
            f"storage://{dst_cluster}/{current_user}/copy-dst/{str(uuid.uuid4())}"
        )

        result = cli_runner(
            ["neuro", "data-transfer", src_path, dst_path],
        )
        assert result.returncode == 0, result

        del_result = cli_runner(
            ["neuro", "rm", "-r", src_path],
        )
        assert del_result.returncode == 0, result

    with switch_cluster(dst_cluster):
        result = cli_runner(
            ["neuro", "ls", dst_path],
        )
        assert result.returncode == 0, result

        del_result = cli_runner(["neuro", "rm", "-r", dst_path])
        assert del_result.returncode == 0, result



"""Module for common functionality and key abstractions related to data copy"""
import abc
import logging
import os
import re
from dataclasses import dataclass
from enum import Flag, auto
from functools import cached_property
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Optional, Tuple, Union

from neuro_sdk import Client
from yarl import URL


logger = logging.getLogger(__name__)


class ArchiveType(int, Flag):  # type: ignore
    """Int Flag for archive types

    Supports fuzzy checks:
    >>> assert ArchiveType.TAR_GZ == ArchiveType.TAR
    >>> assert (ArchiveType.GZ | ArchiveType.ZIP) == ArchiveType.SUPPORTED
    """

    TAR_PLAIN = auto()
    TAR_GZ = auto()
    TAR_BZ = auto()
    GZ = auto()
    ZIP = auto()
    TAR = TAR_PLAIN | TAR_GZ | TAR_BZ
    SUPPORTED = TAR | GZ | ZIP
    UNSUPPORTED = ~(SUPPORTED)

    @staticmethod
    def get_extensions_for_type(type: "ArchiveType") -> List[str]:
        """Get list of file extensions, that correspond
        to the provided archive type"""
        return [
            ext
            for ext, type_ in ArchiveType.get_extension_mapping().items()
            if type_ == type
        ]

    @staticmethod
    def get_extension_mapping() -> Dict[str, "ArchiveType"]:
        """Get mapping from file extension to ArchiveType"""
        return {
            ".tar.gz": ArchiveType.TAR_GZ,
            ".tgz": ArchiveType.TAR_GZ,
            ".tar.bz2": ArchiveType.TAR_BZ,
            ".bz2": ArchiveType.TAR_BZ,
            ".tbz": ArchiveType.TAR_BZ,
            ".tar": ArchiveType.TAR_PLAIN,
            ".gz": ArchiveType.GZ,
            ".zip": ArchiveType.ZIP,
        }

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ArchiveType):
            return bool(int(self) & int(other))
        return False

    def __hash__(self) -> int:
        return hash(int(self))

    @staticmethod
    def get_type(archive: Path) -> "ArchiveType":
        """Determine archive type from file extension"""
        suffixes = archive.suffixes[-2:]  # keep only at most 2 suffixes
        if not suffixes:
            return ArchiveType.UNSUPPORTED
        if "".join(suffixes) in ArchiveType.get_extension_mapping():
            # match longest possible suffix first
            return ArchiveType.get_extension_mapping()["".join(suffixes)]
        else:
            # try to match last suffix
            return ArchiveType.get_extension_mapping().get(
                suffixes[-1], ArchiveType.UNSUPPORTED
            )


class DataUrlType(int, Flag):  # type: ignore
    """Enum type for handling source/destination types

    Supports comparisons between particular type and its category:
    >>> assert UrlType.GCS == UrlType.CLOUD
    >>> assert UrlType.LOCAL == (~UrlType.UNSUPPORTED)
    """

    LOCAL_FS = auto()
    AZURE = auto()
    S3 = auto()
    GCS = auto()
    HTTP = auto()
    HTTPS = auto()
    WEB = HTTP | HTTPS
    CLOUD = AZURE | S3 | GCS | WEB
    STORAGE = auto()
    DISK = auto()
    PLATFORM = STORAGE | DISK
    COPY_UNSUPPORTED = ~(LOCAL_FS | CLOUD | PLATFORM)
    COPY_SUPPORTED = ~COPY_UNSUPPORTED

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, DataUrlType):
            return bool(self & other)
        return False

    def __hash__(self) -> int:
        return hash(int(self))

    @staticmethod
    def get_scheme_mapping() -> Dict[str, "DataUrlType"]:
        """Get supported mapping of url schema to UrlType"""
        return {
            "": DataUrlType.LOCAL_FS,
            "s3": DataUrlType.S3,
            "gs": DataUrlType.GCS,
            "azure+https": DataUrlType.AZURE,
            "storage": DataUrlType.STORAGE,
            "disk": DataUrlType.DISK,
            "http": DataUrlType.HTTP,
            "https": DataUrlType.HTTPS,
        }

    @staticmethod
    def get_type(url: Union[str, URL]) -> "DataUrlType":
        """Detect UrlType by checking url schema"""
        scheme_mapping = DataUrlType.get_scheme_mapping()
        if isinstance(url, URL):
            url_scheme = url.scheme
        else:
            url_scheme = URL(url).scheme
        return scheme_mapping.get(url_scheme, DataUrlType.COPY_UNSUPPORTED)


class Copier(metaclass=abc.ABCMeta):
    """Base interface for copying data between a variety of sources"""

    def __init__(self, source: "Resource", destination: "Resource") -> None:
        self.source = source
        self.destination = destination
        self._ensure_can_execute()

    def _ensure_can_execute(self) -> None:
        """Raise error if copy cannot be executed"""
        raise NotImplementedError

    async def perform_copy(self) -> "Resource":
        """Copy data from self.source to self.destination
        and return path to the copied resource"""
        raise NotImplementedError


@dataclass(frozen=True)
class Resource:
    """Represents a resource, pointed to by a URL"""

    url: URL
    _client: Optional[Client] = None  # only present for platform resources

    @cached_property
    def data_copy_supported(self) -> bool:
        result = self.data_url_type == DataUrlType.COPY_SUPPORTED
        logger.debug(f"Data copy supported for {self.url}: {result}")
        return result

    @cached_property
    def data_url_type(self) -> DataUrlType:
        result = DataUrlType.get_type(self.url)
        logger.debug(f"Data url type of {self.url}: {result.name}")
        return result

    @cached_property
    def archive_type(self) -> ArchiveType:
        result = ArchiveType.get_type(archive=Path(self.url.path))
        logger.debug(f"Archive type of {self.url}: {result.name}")
        return result

    @cached_property
    def filename(self) -> Optional[str]:
        """Filename to which the url is pointing

        For local urls, this is obtained from os.path.split
        For platform urls this is obtained by detecting cluster/org/project name,
          if the url points to a project/org/cluster (without trailing slash),
          the filename will be None.
        For all other urls - the last part of self.url.parts is returned
        """
        if self.data_url_type == DataUrlType.LOCAL_FS:
            # use pathlib
            _, tail = os.path.split(self.as_path())
            result = tail if tail else None

        elif self.data_url_type in (DataUrlType.DISK, DataUrlType.STORAGE):
            assert self._client is not None
            normalized_url = self._client.parse.normalize_uri(
                self.strip_mount_mode_flag().url
            )
            parts = normalized_url.parts
            # during normaliziation at least current project is present
            assert len(parts) >= 2
            cluster_name = normalized_url.host
            assert cluster_name is not None
            if parts[1] in self._client.config.cluster_orgs:
                # url of type {schema}://{cluster_name}/{org}/{project}[rest]
                logger.debug(f"{self.url} (normalized: {normalized_url}) has org")
                assert len(parts) >= 3
                parts_to_skip = 3
            else:
                # url of type {schema}://{cluster_name}/{project}[rest]
                logger.debug(f"{self.url} (normalized: {normalized_url}) has no org")
                parts_to_skip = 2
            if self.data_url_type == DataUrlType.DISK:
                # url of type
                # {schema}://{cluster_name}[/{org}]/{project}/{disk-id}[rest]
                parts_to_skip += 1
            parts = parts[parts_to_skip:]
            if parts:
                result = parts[-1] if parts[-1] else None
            else:
                result = None
        else:
            parts = self.url.parts
            if parts:
                result = parts[-1] if parts[-1] else None
            else:
                result = None
        logger.debug(f"Filename of {self.url} ({self.data_url_type.name}): {result}")
        return result

    @cached_property
    def disk_id_and_path(self) -> Tuple[str, Optional[str]]:
        """For disk resources - full id of the disk
        (disk://{cluster}[/org]/project/id) and the path on disk

        Fails if self._client is None.
        """
        assert self.data_url_type == DataUrlType.DISK
        assert self._client is not None
        normalized_url = self._client.parse.normalize_uri(
            self.strip_mount_mode_flag().url
        )
        parts = normalized_url.parts
        # during normaliziation at least current project and disk id is present
        assert len(parts) >= 3
        cluster_name = normalized_url.host
        assert cluster_name is not None
        if parts[1] in self._client.config.cluster_orgs:
            # url of type {schema}://{cluster_name}/{org}/{project}[rest]
            logger.debug(f"{self.url} (normalized: {normalized_url}) has org")
            assert len(parts) >= 3
            disk_name_len = 4
        else:
            logger.debug(f"{self.url} (normalized: {normalized_url}) has no org")
            disk_name_len = 3
        org_project_name = "/".join(parts[1:disk_name_len])
        full_disk_id = f"disk://{cluster_name}/{org_project_name}"
        if parts[disk_name_len:]:
            path_on_disk = "/" + "/".join(parts[disk_name_len:])
        else:
            path_on_disk = None
        logger.debug(f"{self.url} {full_disk_id=} {path_on_disk=}")
        return full_disk_id, path_on_disk

    @staticmethod
    def parse(url: str, client: Client) -> "Resource":
        """Parse resource spec, normalizing platform urls if needed"""
        parsed_url = URL(url)
        scheme = parsed_url.scheme
        if scheme in ("storage", "disk"):
            return Resource(
                url=client.parse.normalize_uri(parsed_url),
                _client=client,
            )
        return Resource(url=parsed_url)

    @staticmethod
    def from_path(path: Path) -> "Resource":
        return Resource(URL(str(path)))

    @staticmethod
    def from_str(url: str, *, _client: Optional[Client] = None) -> "Resource":
        """Create a resource from the string url, without normalization"""
        return Resource(URL(url), _client=_client)

    def as_path(self) -> Path:
        return Path(self.url.path)

    def as_str(self) -> str:
        return str(self.url)

    def __str__(self) -> str:
        return self.as_str()

    def strip_filename(self) -> "Resource":
        """
        Return a Resource pointing to the parent folder
        if the url points to a file
        """
        if self.filename is None:
            logger.debug(f"{self.url} has no fileneme")
            return self
        result = Resource.from_str(
            re.sub(pattern=f"{self.filename}$", repl="", string=self.as_str()),
            _client=self._client,
        )
        logger.debug(f"Stripped filename from {self.url} to {result.url}")
        return result

    @cached_property
    def mode_flag(self) -> Optional[str]:
        """Get :ro or :rw flag from platform urls"""
        if self.data_url_type not in (DataUrlType.DISK, DataUrlType.STORAGE):
            logger.debug(f"{self.url} ({self.data_url_type=}) has no mode flag.")
            return None
        match = re.search(pattern=f":(rw|ro)$", string=self.as_str())
        if match:
            result = match.group(1)
            logger.debug(f"Mode flag of {self.url}: {result}")
            return result
        logger.debug(f"{self.url} ({self.data_url_type=}) has no mode flag.")
        return None

    def strip_mount_mode_flag(self) -> "Resource":
        if self.data_url_type in (DataUrlType.DISK, DataUrlType.STORAGE):
            return Resource.from_str(
                re.sub(pattern="(:ro|:rw)$", repl="", string=self.as_str()),
                _client=self._client,
            )
        return self


def ensure_folder_exists(local_resource: Resource) -> None:
    """Ensure the folder (in case of the file resource - parent folder) exists"""
    folder_name, _ = os.path.split(local_resource.as_str())
    logger.info(f"Creating folder for {folder_name}")
    Path(folder_name).mkdir(exist_ok=True, parents=True)


def get_default_preset(neuro_client: Client) -> str:
    """Get default preset name via Neu.ro client"""
    return next(iter(neuro_client.presets.keys()))


def provide_temp_dir(
    dir: Path = Path.home() / ".neuro-tmp",
) -> TemporaryDirectory:  # type: ignore
    """Provide temp directory"""
    dir.mkdir(exist_ok=True, parents=True)
    return TemporaryDirectory(dir=dir)



import logging
import uuid
from typing import List

from ..conftest import (
    CLOUD_SOURCE_PREFIXES,
    PLATFORM_DESTINATION_PREFIXES,
    get_tested_archive_types,
)
from .resources import CopyTestConfig, DataTestResource


logger = logging.getLogger(__name__)


def generate_cloud_to_platform_copy_configs() -> List[CopyTestConfig]:
    test_configs: List[CopyTestConfig] = []
    archive_types = get_tested_archive_types()
    run_uuid = uuid.uuid4().hex
    for schema, cloud_url in CLOUD_SOURCE_PREFIXES.items():
        dir_copy_should_fail = schema in ("http", "https")
        dir_copy_fail_reason = "Copy from HTTP(S) dir is unsupported"
        for platform_schema, platform_prefix in PLATFORM_DESTINATION_PREFIXES.items():
            cloud_folder = DataTestResource(
                schema=schema,
                url=f"{cloud_url}/",
                is_archive=False,
                file_extension=None,
            )
            platform_copy_folder = DataTestResource(
                schema=platform_schema,
                url=f"{platform_prefix}/{run_uuid}/copy/{schema}/folder/",
                is_archive=False,
                file_extension=None,
            )

            for ext in archive_types:
                # test for copy of remote archive into local one
                cloud_archive = DataTestResource(
                    schema=schema,
                    url=f"{cloud_url}/file{ext}",
                    is_archive=True,
                    file_extension=ext,
                )
                platform_archive = DataTestResource(
                    schema=platform_schema,
                    url=f"{platform_prefix}/{run_uuid}/compress/{schema}/file{ext}",
                    is_archive=True,
                    file_extension=ext,
                )

                test_configs.append(
                    CopyTestConfig(
                        source=cloud_archive,
                        destination=platform_archive,
                        compress_flag=True,
                    )
                )
                # test for extraction of cloud archive into local dir
                platform_extract_folder = DataTestResource(
                    schema=platform_schema,
                    url=f"{platform_prefix}/{run_uuid}/extract/{schema}/{ext}/",
                    is_archive=False,
                )
                test_configs.append(
                    CopyTestConfig(
                        source=cloud_archive,
                        destination=platform_extract_folder,
                        extract_flag=True,
                    )
                )

                # test for compression of cloud folder into local archive
                platform_compressed_archive = DataTestResource(
                    schema=platform_schema,
                    url=f"{platform_prefix}/{run_uuid}/compress/{schema}/file{ext}",
                    is_archive=True,
                    file_extension=ext,
                )

                test_configs.append(
                    CopyTestConfig(
                        source=cloud_folder,
                        destination=platform_compressed_archive,
                        compress_flag=True,
                        should_fail=dir_copy_should_fail,
                        fail_reason=dir_copy_fail_reason,
                    )
                )
            if archive_types:
                # use the values for source and destination files
                # from last loop iteration
                test_configs.append(
                    CopyTestConfig(
                        source=cloud_archive,
                        destination=platform_archive,
                    )
                )
            test_configs.append(
                CopyTestConfig(
                    source=cloud_folder,
                    destination=platform_copy_folder,
                    should_fail=dir_copy_should_fail,
                    fail_reason=dir_copy_fail_reason,
                )
            )

    logger.info(f"Generated cloud to platform tests: {test_configs}")
    return test_configs



"""Module for copying files on local filesystem"""
import os
from pathlib import Path

from ..utils import CLIRunner
from .common import Copier, DataUrlType, Resource


class LocalFSCopier(Copier, CLIRunner):
    """Copier implementation for local file system operations"""

    def _ensure_can_execute(self) -> None:
        if not (
            self.source.data_url_type == DataUrlType.LOCAL_FS
            and self.destination.data_url_type == DataUrlType.LOCAL_FS
        ):
            raise ValueError("Only local filesystem is supported")

    async def perform_copy(self) -> Resource:
        """Perform copy through running rclone and return the url to destinaton"""
        destination_path = Path(self.destination.url.path)
        destination_parent_folder, _ = os.path.split(destination_path)
        Path(destination_parent_folder).mkdir(exist_ok=True, parents=True)
        command = "rclone"
        args = [
            "copyto",  # TODO: investigate usage of 'sync' for potential speedup.
            "--checkers=16",  # https://rclone.org/docs/#checkers-n , default is 8
            "--transfers=8",  # https://rclone.org/docs/#transfers-n , default is 4.
            "--verbose=1",  # default is 0, set 2 for debug
            self.source.as_str(),
            self.destination.as_str(),
        ]
        await self.run_command(command=command, args=args)
        return self.destination



"""Module for archive management operations (compression and extraction)"""
import abc
import logging

from ..utils import CLIRunner
from .common import ArchiveType, Resource, ensure_folder_exists


logger = logging.getLogger(__name__)


class ArchiveManager(metaclass=abc.ABCMeta):
    """Interface for archive management"""

    @abc.abstractmethod
    async def compress(self, source: Resource, destination: Resource) -> Resource:
        """Compress source into destination"""
        raise NotImplementedError

    @abc.abstractmethod
    async def extract(self, source: Resource, destination: Resource) -> Resource:
        """Extract source into destination"""
        raise NotImplementedError


class TarManager(ArchiveManager, CLIRunner):
    """Utility class for handling tar archives"""

    async def compress(self, source: Resource, destination: Resource) -> Resource:
        """Compress source into destination using tar command"""
        command = "tar"
        if not destination.archive_type == ArchiveType.TAR:
            raise ValueError(
                f"Can't compress into {destination.url} with TarManager: "
                f"unsupported archive type {destination.archive_type.name}. "
                f"Supported types: "
                f"{ArchiveType.get_extensions_for_type(ArchiveType.TAR)}"
            )
        mapping = {
            ArchiveType.TAR_GZ: "zcvf",
            ArchiveType.TAR_BZ: "jcvf",
            ArchiveType.TAR_PLAIN: "cvf",
        }
        subcommand = mapping[destination.archive_type]
        args = [
            subcommand,
            str(destination),
            # f"--exclude={destination.filename}",
            str(source),
        ]
        await self.run_command(command=command, args=args)
        return destination

    async def extract(self, source: Resource, destination: Resource) -> Resource:
        """Extract source into destination using tar command"""
        command = "tar"
        if not source.archive_type == ArchiveType.TAR:
            raise ValueError(
                f"Can't extract {source} with TarManager: "
                f"unsupported archive type {source.archive_type.name}. "
                f"Supported types: "
                f"{ArchiveType.get_extensions_for_type(ArchiveType.TAR)}"
            )
        mapping = {
            ArchiveType.TAR_GZ: "zxvf",
            ArchiveType.TAR_BZ: "jxvf",
            ArchiveType.TAR_PLAIN: "xvf",
        }
        subcommand = mapping[source.archive_type]
        args = [subcommand, source.as_str(), f"-C", destination.as_str()]
        destination.as_path().mkdir(exist_ok=True, parents=True)
        await self.run_command(command=command, args=args)
        return destination


class GzipManager(ArchiveManager, CLIRunner):
    """Utility class for handling gzip archives"""

    async def compress(self, source: Resource, destination: Resource) -> Resource:
        """Compress source into destination using gzip command"""
        command = "gzip"
        if not destination.archive_type == ArchiveType.GZ:
            raise ValueError(
                f"Can't compress into {destination} with GzipManager: "
                f"unsupported archive type {destination.archive_type.name}. "
                f"Supported types: "
                f"{ArchiveType.get_extensions_for_type(ArchiveType.GZ)}"
            )
        if source.filename is None:
            raise ValueError(
                "gzip does not support folder compression, "
                "use .tar.gz extension instead."
            )
        args = ["-rkvf", source.as_str()]
        await self.run_command(command=command, args=args)
        # gzip does not support setting destination
        temp_destination = source.as_str() + ".gz"
        # TODO: add support for non-unix OS
        await self.run_command("mv", ["-v", temp_destination, destination.as_str()])
        return destination

    async def extract(self, source: Resource, destination: Resource) -> Resource:
        """Extract source into destination using gunzip command"""
        command = "gunzip"
        if not source.archive_type == ArchiveType.GZ:
            raise ValueError(
                f"Can't extract {source} with GzipManager: "
                f"unsupported archive type {source.archive_type.name}. "
                f"Supported types: "
                f"{ArchiveType.get_extensions_for_type(ArchiveType.GZ)}"
            )
        args = ["--keep", source.as_str()]
        await self.run_command(command=command, args=args)
        temp_destination = str(
            source.as_path().with_suffix("")
        )  # gzip extracts inplace
        await self.run_command("mv", ["-v", temp_destination, destination.as_str()])
        return destination


class ZipManager(ArchiveManager, CLIRunner):
    """Utility class for handling zip archives"""

    async def compress(self, source: Resource, destination: Resource) -> Resource:
        """Compress source into destination using zip command"""
        command = "zip"
        if not destination.archive_type == ArchiveType.ZIP:
            raise ValueError(
                f"Can't compress into {destination} with ZipManager: "
                f"unsupported archive type {destination.archive_type.name}. "
                f"Supported types: "
                f"{ArchiveType.get_extensions_for_type(ArchiveType.ZIP)}"
            )
        # check if works as expected
        args = ["-rv", destination.as_str(), source.as_str()]
        await self.run_command(command=command, args=args)
        return destination

    async def extract(self, source: Resource, destination: Resource) -> Resource:
        """Extract source into destination using unzip command"""
        command = "unzip"
        if not source.archive_type == ArchiveType.ZIP:
            raise ValueError(
                f"Can't extract {source} with ZipManager: "
                f"unsupported archive type {source.archive_type.name}. "
                f"Supported types: "
                f"{ArchiveType.get_extensions_for_type(ArchiveType.ZIP)}"
            )
        args = ["-o", source.as_str(), "-d", destination.as_str()]
        destination.as_path().mkdir(exist_ok=True, parents=True)
        await self.run_command(command=command, args=args)
        return destination


def _get_archive_manager(archive: Resource) -> ArchiveManager:
    """Resolve appropriate archive manager"""
    mapping = {
        ArchiveType.TAR: TarManager(),
        ArchiveType.GZ: GzipManager(),
        ArchiveType.ZIP: ZipManager(),
    }
    archive_type = ArchiveType.get_type(archive.as_path())
    if archive_type == ArchiveType.UNSUPPORTED:
        supported_extensions = list(ArchiveType.get_extension_mapping())
        raise ValueError(
            f"Unsupported archive type for file {archive}, "
            f"supported types are {supported_extensions}"
        )
    return next(manager for type, manager in mapping.items() if type == archive_type)


async def copy(source: Resource, destination: Resource) -> Resource:
    """Copy source into destination"""
    command = "cp"
    args = [source.as_str(), destination.as_str()]
    runner = CLIRunner()
    await runner.run_command(command=command, args=args)
    return destination


async def compress(source: Resource, destination: Resource) -> Resource:
    """Compress source into destination while
    inferring arhive type from destination"""
    ensure_folder_exists(destination)
    if source.filename is not None and destination.filename is not None:
        both_archives = ArchiveType.UNSUPPORTED not in (
            source.archive_type,
            destination.archive_type,
        )
        same_type = source.archive_type == destination.archive_type
        if both_archives and same_type:
            logger.info(
                "Skipping compression step - "
                "source is already archive of the same type"
            )
            return await copy(source=source, destination=destination)

    manager_implementation = _get_archive_manager(destination)
    logger.debug(
        f"Compressing {source} into {destination} "
        f"with {manager_implementation.__class__.__name__}"
    )
    return await manager_implementation.compress(source=source, destination=destination)


async def extract(source: Resource, destination: Resource) -> Resource:
    """Extract source into destination while
    inferring arhive type from source"""
    ensure_folder_exists(destination)

    manager_implementation = _get_archive_manager(source)
    logger.debug(
        f"Extracting {source} into {destination} "
        f"with {manager_implementation.__class__.__name__}"
    )
    return await manager_implementation.extract(source=source, destination=destination)



import json
from pathlib import Path
from unittest import mock

import pytest

from .conftest import CLIRunner


@pytest.mark.smoke
def test_config_save_registry_auth_locally(cli_runner: CLIRunner) -> None:
    result = cli_runner(
        ["neuro-extras", "config", "save-registry-auth", ".docker.json"]
    )
    assert result.returncode == 0, result

    with Path(".docker.json").open() as f:
        payload = json.load(f)

    assert payload == {"auths": mock.ANY}



# Python on Windows doesn't expose these constants in os module

EX_CANTCREAT = 73
EX_CONFIG = 78
EX_DATAERR = 65
EX_IOERR = 74
EX_NOHOST = 68
EX_NOINPUT = 66
EX_NOPERM = 77
EX_NOUSER = 67
EX_OK = 0
EX_OSERR = 71
EX_OSFILE = 72
EX_PROTOCOL = 76
EX_SOFTWARE = 70
EX_TEMPFAIL = 75
EX_UNAVAILABLE = 69
EX_USAGE = 64

# os module has no this constant but timeout posix command exits with 124,
# see 'man 1 timeout'
EX_TIMEOUT = 124
EX_PLATFORMERROR = 125  # neuro platform misfunctioning



from pathlib import Path
from typing import List

from ..conftest import get_tested_archive_types
from .resources import TEMPDIR_PREFIX, CopyTestConfig, DataTestResource


def generate_local_to_local_copy_configs() -> List[CopyTestConfig]:
    assets_root = (Path(__file__).parent.parent.parent / "assets" / "data").resolve()
    archive_types = get_tested_archive_types()
    test_configs: List[CopyTestConfig] = []
    plain_copy = [
        CopyTestConfig(
            source=DataTestResource(
                schema="local",
                url=str(assets_root / f"file{ext}"),
                file_extension=ext,
                is_archive=True,
            ),
            destination=DataTestResource(
                schema="local",
                url=f"{TEMPDIR_PREFIX}/copy/local/file{ext}",
                file_extension=ext,
                is_archive=True,
            ),
        )
        for ext in archive_types
    ]
    test_configs += plain_copy
    return test_configs
